; Joust game
; Note: collisions can only be detected by moving actors

(define root (new-SoSeparator))
; always ref the root
(-> root 'ref)
(define viewer (examiner root))
(-> viewer 'setViewing 0)

(load "game.scm")

(define *speed* (new-SbTime (/ 1 5)))
(define *playfield-box* (new-SbBox3f -1 -1 -1 1 1 1))
(define *playfield-faces* #(1 1 -0.7 -1 0 0))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; actor types

; Barriers just sit there
(define (new-Barrier root pos length)
  (define my-root (new-SoSeparator))
  (define cube (new-SoCube))
  (-> (-> cube 'height) 'setValue 0.03)
  (-> (-> cube 'depth) 'setValue 0.1)
  (-> (-> cube 'width) 'setValue length)
  (addChildren my-root white cube)

  (define self (new-Actor root my-root pos))
  (send self 'addTag 'barrier)
  
  self)

; Eggs fall to the ground and collide with boundaries, Barriers, and eachother
(define (new-Egg root pos color)
  (define my-root (new-SoSeparator))
  (define sphere (new-SoSphere))
  (-> (-> sphere 'radius) 'setValue 0.05)
  (addChildren my-root color sphere)

  (define self (new-Actor root my-root pos))
  (send self 'addTag 'heavy)
  (send self 'addTag 'egg)

  (send self 'addBoundsCheck *playfield-box* *playfield-faces*)
  (send self 'addCollisionHandler viewer '(egg)
	(lambda (self other)
	  (elastic-collision self other)
	  ))
  (send self 'addCollisionHandler viewer '(barrier)
	(lambda (self other)
	  ;(elastic-collision self other 1 0)
          (define v (send self 'getVelocity))
          (actor-separate self other viewer)
          (send self 'setVelocity (vector-scale -1 v))
	  ))

  self)

; Horses collide with barriers and the playfield boundaries
(define (new-Horse root pos color)
  (define my-root (new-SoSeparator))
  (define cone (new-SoCone))
  (-> (-> cone 'bottomRadius) 'setValue 0.1)
  (-> (-> cone 'height) 'setValue 0.15)
  (addChildren my-root color cone)

  (define self (new-Actor root my-root pos))
  (send self 'addTag 'heavy)

  (send self 'addBoundsCheck *playfield-box* *playfield-faces*)
  (send self 'addCollisionHandler viewer '(barrier)
	(lambda (self other)
	  ;(elastic-collision self other 1 0)
          (define v (send self 'getVelocity))
          (actor-separate self other viewer)
          (send self 'setVelocity (vector-scale -1 v))
	  ))
  
  self)

; Player is a Horse that collides with Enemies and Eggs
(define (new-Player horse)
  (define wounded #f)

  (define (self message)
    (cond ((eq? message 'die)
	   (lambda (self)
	     (-> (send horse 'shape) 'replaceChild 0 cyan)
	     (set! wounded #t)
	     (set! *keyboard-active* #f)
	     (define (alarm-cb user-data dummy)
	       (-> (send horse 'shape) 'replaceChild 0 blue)
	       (set! wounded #f)
	       (set! *keyboard-active* #t))
	     (define alarm (sensor new-SoAlarmSensor alarm-cb))
	     (-> alarm 'setTimeFromNow (new-SbTime 5.0))
	     (-> alarm 'schedule)
	     ))

	  (else (horse message))))

  (send self 'addCollisionHandler viewer '(enemy)
	(lambda (self other)
	  (elastic-collision self other)
	  (define slack 0.02)
	  (define delta (- (vector-ref (send self 'getPosition) 1)
			   (vector-ref (send other 'getPosition) 1)))
	  (cond ((and (not wounded) (> delta slack))
		 (define egg (new-Egg root (send other 'getPosition) white))
		 (send egg 'setVelocity (send other 'getVelocity))
		 (send other 'die))
		((and (not wounded) (> (- delta) slack))
		 (send self 'die)
		 (add-enemy)
		 ))
	  ))

  (send self 'addCollisionHandler viewer '(egg)
	(lambda (self other)
	  (elastic-collision self other)
	  (cond ((not wounded)
		 (send other 'die)
		 (if (< (num-enemies) 2) (add-enemy))
		 ))
	  ))

  self)

; Enemy is an autonomous Horse that collides with other Enemies and Eggs
(define (new-Enemy horse)
  (define moving-left (< (vector-ref (send horse 'getVelocity) 0) 0))
  (define (brain-cb user-data dummy)
    (define p (send horse 'getPosition))
    (define pp (send player 'getPosition))
    (if (= (random 2) 1)
	; flap to catch up with player
	(if (< (vector-ref p 1) (vector-ref pp 1))
	    (flap horse))
	; flap to stay in the top half
	(if (< (vector-ref p 1) (random-float)) 
	    (flap horse)))
    (if (= (random 2) 1)
	(if (= (random 2) 1)
	    ; face the player
	    (set! moving-left (< (vector-ref pp 0) (vector-ref p 0))))
	; accelerate
	(if moving-left
	    (move-left horse)
	    (move-right horse)))
    #f)
  (define brain (sensor new-SoTimerSensor brain-cb))
  (-> brain 'setInterval *speed*)
  (-> brain 'schedule)
  
  (define (self message)
    (cond ((eq? message 'die)
	   (lambda (self)
	     (-> brain 'unschedule)
	     ((horse 'removeAllTags) self)
	     (send horse 'die)
	     ))

	  (else (horse message))))
  (send self 'addTag 'enemy)

  (send self 'addCollisionHandler viewer '(enemy egg)
	(lambda (self other)
	  (elastic-collision self other)
	  ))

  self)

(define (num-enemies)
  (length (send *actor-tags* 'withTag 'enemy)))
(define (add-enemy)
  (if (< (num-enemies) 4)
      (new-Enemy (new-Horse root #(-0.5 0.35 0) red))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; create actors
(new-Barrier root #(-0.5 0.3 0) 0.45)

(define player (new-Player (new-Horse root #(0 0 0) blue)))

(cond (#t
(new-Enemy (new-Horse root #(-0.5 0.35 0) red))
))

(cond (#t
(new-Enemy (new-Horse root #(0 0.35 0) red))
))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; gravity pulls down heavy actors
(define (gravity-cb user-data dummy)
  (for-each
   (lambda (actor)
     (send actor 'setVelocity (vector-add (send actor 'getVelocity)
					  #(0 -0.2 0))))
   (send *actor-tags* 'withTag 'heavy)))
(define gravity (sensor new-SoTimerSensor gravity-cb))
(-> gravity 'setInterval *speed*)
(-> gravity 'schedule)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; movement routines
(define (move-left actor)
  (define v (send actor 'getVelocity))
  (if (> (vector-ref v 0) -0.4)
      (send actor 'setVelocity (vector-add v #(-0.2 0 0))))
  )
(define (move-right actor)
  (define v (send actor 'getVelocity))
  (if (< (vector-ref v 0) 0.4)
      (send actor 'setVelocity (vector-add v #(0.2 0 0))))
  )
(define (flap actor)
  (define v (send actor 'getVelocity))
  (if (< (vector-ref v 1) 0.8)
      (send actor 'setVelocity (vector-add v #(0 0.4 0))))
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; input
(keyboard-handler root
 (lambda (event)
   (cond (*keyboard-active*
	  (if (= 1 (SO_KEY_PRESS_EVENT event LEFT_ARROW))
	      (move-left player))
	  (if (= 1 (SO_KEY_PRESS_EVENT event RIGHT_ARROW))
	      (move-right player))
	  (if (= 1 (SO_KEY_PRESS_EVENT event UP_ARROW))
	      (flap player))
	  ))
   ))
